/*
#
#  Copyright (c) 2010    David Albert Bagley, bagleyd AT verizon.net
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "usable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Coded after looking at Jaap's Puck page
# OK for puck puzzles but really slow for general when -b is used
# Runtime goes up exponentially with level
*/

#include <stdio.h>
#include <stdlib.h>
#define LEVELS 14
#define WEDGES 12
#define BANDS 2

#define ringAdd(i, k) ((i + k + wedges) % wedges)
typedef struct _pucktype {
	int wedge;
	int dir;
} pucktype;

/* Not much reason to make this dynamic as it really will not
  make any more efficient as it slows down too much at its present limit. */
static pucktype puck[WEDGES][BANDS];
static int fMoves[LEVELS];
static int rMoves[2][LEVELS];

static int levels = -1;
static int wedges = -1;
static int bands = 1;
static int match = -1;
static int half = -1;
static int start = -1;
static int space = 1;
static int radius = 0;
static int order = 0;
static int whole = 0;
static int nomore = 0;

void initPuck()
{
	int i, j;

	for (i = 0; i < wedges; i++) {
		for (j = 0; j < bands; j++) {
			puck[i][j].wedge = i;
			puck[i][j].dir = 0;
		}
	}
}

void initMoves()
{
	int level;

	for (level = 0; level < levels; level++) {
		fMoves[level] = -1;
		rMoves[0][level] = 0;
		rMoves[1][level] = 0;
	}
}

void printPuck()
{
	int i, j;

	if (bands == 1)
		(void) printf("puck\n");
	else
		(void) printf("ball\n");
	for (i = 0; i < wedges; i++) {
		for (j = 0; j < bands; j++) {
			(void) printf(" %d %d ",
				puck[i][j].wedge, puck[i][j].dir);
		}
		(void) printf("\n");
	}
}

void printMoves()
{
	int i;

	(void) printf("moves\n");
	for (i = 0; i < levels; i++) {
		if (fMoves[i] == -1)
			break;
		if (bands > 1)
			(void) printf(" ");
		(void) printf(" %d", fMoves[i]);
		if (bands > 1)
			(void) printf(" %d %d", rMoves[0][i], rMoves[1][i]);
	}
	(void) printf("\n");
}

void f(int o)
{
	pucktype temp[WEDGES / 2][BANDS];
	int i, j;

	for (i = 0; i < half; i++) {
		for (j = 0; j < bands; j++) {
			temp[i][j].wedge = puck[ringAdd(i, o)][j].wedge;
			temp[i][j].dir = puck[ringAdd(i, o)][j].dir;
		}
	}
	for (i = 0; i < half; i++) {
		for (j = 0; j < bands; j++) {
			puck[ringAdd(o, half - 1 - i)][bands - 1 - j].wedge =
				temp[i][j].wedge;
			puck[ringAdd(o, half - 1 - i)][bands - 1 - j].dir =
				(temp[i][j].dir == 0) ? 1 : 0;
		}
	}
}

void r(int b, int o)
{
	pucktype temp[WEDGES];
	int i;

	for (i = 0; i < wedges; i++) {
		temp[i].wedge = puck[i][b].wedge;
		temp[i].dir = puck[i][b].dir;
	}
	for (i = 0; i < wedges; i++) {
		puck[i][b].wedge = temp[ringAdd(i, o)].wedge;
		puck[i][b].dir = temp[ringAdd(i, o)].dir;
	}
}

/* This finds n wedges in a row that are not messed up */
int findOut(int n)
{
	int count = 0, maxCount = 0, i, j, k;

	if (whole && bands > 1) {
		int k1 = -1, k2 = -1, count1 = 0;

		for (i = 0; i < wedges + n + space - 1; i++) {
			k = i % wedges;
			for (j = 0; j < bands / 2; j++) {
				if (puck[k][j].wedge == k &&
				    puck[k][j].dir == 0 &&
				    puck[k][bands - 1 - j].wedge == k &&
				    puck[k][bands - 1 - j].dir == 0) {
					count++;
					if (maxCount < count)
						maxCount = count;
				} else {
					
					if (space == 2) {
						if (count1 == 1)
							count = 0;
						count1 = 1;
					} else {
						count = 0;
					}
					if (order)
						break;
					if (k1 < 0 || k1 == k)
						k1 = k;
					else
						k2 = k;
				}
			}
		}
		if (nomore) {
			if (maxCount == n) {
				if (space == 2) {
					if (((k1 + 1) % wedges != k2) &&
					    ((k2 + 1) % wedges != k1) &&
					    k2 >= 0) {
int i1, i2;
printf("k1 %d, k2 %d\n", k1, k2);
for (i1 = 0; i1 < wedges; i1++) {
for (i2 = i1 + 1; i2 < wedges; i2++) {
	if (puck[i1][0].wedge ==  puck[i2][0].wedge)
		printf("eureka\n");
}
}
						return 1;
					} else
						return 0;
				} else
					return 1;
			}
		} else {
			if (maxCount >= n && maxCount < wedges) {
				if (space == 2) {
					if (((k1 + 1) % wedges != k2) &&
					    ((k2 + 1) % wedges != k1) &&
					    k2 >= 0) {
printf("k1 %d, k2 %d\n", k1, k2);
						return 1;
					} else
						return 0;
				} else
					return 1;
			}
		}
		return 0;
	} else {
#if 0
		int wedge1 = -1, wedge2 = -1;

#endif
			for (j = 0; j < bands; j++) {
//		for (i = 0; i < wedges + n; i++) {
		for (i = 0; i < wedges; i++) {
			k = i % wedges;
				if (puck[k][j].wedge == k &&
				    puck[k][j].dir == 0) {
					count++;
					if (maxCount < count)
						maxCount = count;
				} else {
					count = 0;
#if 0
					if (wedge1 == -1)
						wedge1 = k;
					else
						wedge2 = k;
#endif
					if (order)
						break;
				}
			}
		}
		if (nomore) {
			if (maxCount == n) {
int i1, i2 = 0; i3 = 0;
for (i1 = 0; i1 < wedges; i1++) {
	if (puck[i1][0].dir == 1)
		i2++;
	if (puck[i1][1].dir == 1)
		i3++;
}
if (i2 == 0 || i3 == 0)
	printf("eureka\n");
				return 1;
			}
		} else {
			if (maxCount >= n && maxCount < wedges * bands) {
int i1, i2 = 0; i3 = 0;
for (i1 = 0; i1 < wedges; i1++) {
	if (puck[i1][0].dir == 1)
		i2++;
	if (puck[i1][1].dir == 1)
		i3++;
}
if (i2 == 0 || i3 == 0)
	printf("eureka!\n");
#if 0
				if ((wedge1 + 2) % wedges == wedge2)
#endif
				return 1;
			}
		}
		return 0;
	}
}

/* for non-puck assume for now f and r alternate and
   r is for band 0 and offset is either 1 or -1 */
void tryMoves()
{
	int level;

	initPuck();
	for (level = 0; level < levels; level++) {
		if (fMoves[level] == -1) {
			break;
		}
		f(fMoves[level]);
		if (bands > 1) {
			r(0, rMoves[0][level]);
			r(1, rMoves[1][level]);
		}
	}
	if (findOut(match)) {
		printPuck();
		printMoves();
		(void) printf("--\n");
	}
}

void recurse(int test, int level);

void makeMove(int test, int level)
{
	if (test != 0) {
		tryMoves();
	} else {
		if (level < levels - 1) {
			recurse(1, level + 1);
			recurse(0, level + 1);
		}
	}
}

/* I hope this is BFS */
void recurse(int test, int level)
{
	int i, j, k;

	fMoves[level] = -1;
	rMoves[0][level] = 0;
	rMoves[1][level] = 0;
	if (bands > 1) {
		fMoves[level] = start;
		rMoves[0][level] = 0;
		rMoves[1][level] = 0;
		makeMove(test, level);
		for (j = 1; j <= radius; j++) {
			rMoves[0][level] = j;
			rMoves[1][level] = 0;
			makeMove(test, level);
			for (k = 1; k <= radius; k++) {
				rMoves[1][level] = k;
				makeMove(test, level);
				rMoves[1][level] = -k;
				makeMove(test, level);
			}
			rMoves[0][level] = -j;
			rMoves[1][level] = 0;
			makeMove(test, level);
			for (k = 1; k <= radius; k++) {
				rMoves[1][level] = k;
				makeMove(test, level);
				rMoves[1][level] = -k;
				makeMove(test, level);
			}
		}
	} else {
		for (i = 0; i < wedges; i++) {
			if ((fMoves[level - 1] != i) &&
					(fMoves[level - 1] != ringAdd(i, half))) {
				fMoves[level] = i;
				makeMove(test, level);
			}
		}
	}
	fMoves[level] = -1;
	rMoves[0][level] = 0;
	rMoves[1][level] = 0;
}

void startMoving(int test)
{
	int level = 0, j, k;

	fMoves[level] = start;
	rMoves[0][level] = 0;
	rMoves[1][level] = 0;
	if (bands > 1) {
		for (j = 0; j <= radius; j++) {
			rMoves[0][level] = j;
			rMoves[1][level] = 0;
			makeMove(test, level);
			for (k = 1; k <= radius; k++) {
				rMoves[1][level] = k;
				makeMove(test, level);
				rMoves[1][level] = -k;
				makeMove(test, level);
			}
		}
	} else {
		makeMove(test, level);
	}
	fMoves[level] = -1;
	rMoves[0][level] = 0;
	rMoves[1][level] = 0;
}

void usage()
{
	(void) fprintf(stderr,
                "usage: \"puck [-b] [-h] [-o] [-r radius] [-p space] [-s start] -w wedges -m match -l levels\"\n");
        (void) fprintf(stderr, "\t-b 2 bands (a ball) else 1 band (a puck)\n");
        (void) fprintf(stderr, "\t-h match only whole wedges (affects -m)\n");
        (void) fprintf(stderr, "\t-n no more than match\n");
        (void) fprintf(stderr, "\t-o order from first wedge\n");
        (void) fprintf(stderr, "\t-r radius of how much to search, use with -b option\n");
        (void) fprintf(stderr, "\t-p space between non-matching\n");
        (void) fprintf(stderr, "\t-s starting wedge flip\n");
        (void) fprintf(stderr, "\twedges must be < %d and even\n", WEDGES);
        (void) fprintf(stderr, "\tmatch must be less than wedges (* bands)\n");
        (void) fprintf(stderr, "\tlevels must be < %d\n", LEVELS);
        (void) fprintf(stderr, "\tthere are some more restrictions\n");
        exit(1);
}

int main(int argc, char **argv) {
	argc--;
	argv++;
        while (argc) {
		if (argv[0][0] == '-') {
			switch (argv[0][1]) {
			case '\0':
				(void) fprintf(stderr, "did not like null\n");
				usage();
				break;
			case 'b':
				bands = 2;
				break;
			case 'h':
				whole = 1;
				break;
			case 'n':
				nomore = 1;
				break;
			case 'o':
				order = 1;
				break;
			case 'l':
				if (argc > 1) {
					argc--;
					argv++;
					levels = atoi(argv[0]);
					break;
				} else {
					(void) fprintf(stderr, "did not like l\n");
					usage();
				}
			case 'm':
				if (argc > 1) {
					argc--;
					argv++;
					match = atoi(argv[0]);
					break;
				} else {
					(void) fprintf(stderr, "did not like m\n");
					usage();
				}
			case 'r':
				if (argc > 1) {
					argc--;
					argv++;
					radius = atoi(argv[0]);
					break;
				} else {
					(void) fprintf(stderr, "did not like r\n");
					usage();
				}
			case 'p':
				if (argc > 1) {
					argc--;
					argv++;
					space = atoi(argv[0]);
					break;
				} else {
					(void) fprintf(stderr, "did not like r\n");
					usage();
				}
			case 's':
				if (argc > 1) {
					argc--;
					argv++;
					start = atoi(argv[0]);
					break;
				} else {
					(void) fprintf(stderr, "did not like s\n");
					usage();
				}
			case 'w':
				if (argc > 1) {
					argc--;
					argv++;
					wedges = atoi(argv[0]);
					break;
				} else {
					(void) fprintf(stderr, "did not like w\n");
					usage();
				}
			default:
				(void) fprintf(stderr, "did not like default\n");
				usage();
			}
		}
                argc--;
                argv++;
	}
	if (levels > LEVELS || levels < 0) {
		(void) fprintf(stderr, "did not like levels\n");
		usage();
	}
	if (wedges < 2 || wedges > WEDGES || wedges % 2 == 1) {
		(void) fprintf(stderr, "did not like wedges\n");
		usage();
	}
	if (match < 1) {
		(void) fprintf(stderr, "did not like match\n");
		usage();
	}
	if (bands == 1)
		radius = 0;
	if (radius < 0)
		radius = -radius;
	half = wedges / 2;
	if (radius > half)
		radius = half;
	if (start < 0 || start >= wedges)
		start = 0;
	if (space <= 0 || space > wedges / 2)
		space = 1;
	if (levels % 2 == 1)
		(void) fprintf(stderr, "generally only even number of levels will work\n");
#if !defined( __MSDOS__ ) && !defined( __CYGWIN__ )
	/* This gets rid of the unwanted buffering in UNIX */
	(void) setlinebuf(stdout);
#endif
	initMoves();
	startMoving(1);
	startMoving(0);
#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
