package com.sillycycle.bagleyd.mball;

/*
 * @(#)Mball.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Cursor;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import javax.swing.JFrame;

/**
 * The <code>Mball</code> class extends the
 * <code>MballCanvas</code> for handling the responses to the user
 * for the <code>MballFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class Mball extends MballCanvas implements MouseWheelListener,
		MouseListener, MouseMotionListener, KeyListener {
	private static final long serialVersionUID = 42L;

	public Mball(JFrame frame,
			boolean orient, boolean practice, int wedges, int bands,
			int base, boolean mono, boolean reverse, String names,
			Color c0, Color c1, Color c2, Color c3,
			Color c4, Color c5, Color c6, Color c7,
			Color c8, Color c9, Color c10, Color c11,
			Color borderc, Color fg, Color bg, int delay,
			boolean perspective) {
		this.frame = frame;
		this.orient = orient;
		this.practice = practice;
		this.bands = bands;
		this.wedges = wedges;
		this.base = base;
		this.mono = mono;
		this.reverse = reverse;
		this.wedgeNames = names;
		this.wedgeColor[0] = c0;
		this.wedgeColor[1] = c1;
		this.wedgeColor[2] = c2;
		this.wedgeColor[3] = c3;
		this.wedgeColor[4] = c4;
		this.wedgeColor[5] = c5;
		this.wedgeColor[6] = c6;
		this.wedgeColor[7] = c7;
		this.wedgeColor[8] = c8;
		this.wedgeColor[9] = c9;
		this.wedgeColor[10] = c10;
		this.wedgeColor[11] = c11;
		this.currentForeground = this.borderColor = borderc;
		this.foreground = fg; this.background = bg;
		this.delay = delay;
		this.perspective = perspective;
		firstPaint = true;
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.addMouseWheelListener(this);
		this.addKeyListener(this);
		setCursor(new Cursor(Cursor.HAND_CURSOR));
	}

	public boolean processKey(char key) {
		switch (key) {
		case '>':
		case '.':
			speedUpPuzzle();
			return true;
		case '<':
		case ',':
			slowDownPuzzle();
			return true;
		case '@':
			((MballFrame) frame).toggleSound();
			return true;
		case '?':
			((MballFrame) frame).descriptionHelp();
			return true;
		case '!':
			((MballFrame) frame).featuresHelp();
			return true;
		case '^':
			((MballFrame) frame).referencesHelp();
			return true;
		case 'a':
		case 'A':
			((MballFrame) frame).aboutHelp();
			return true;
		case '*':
			((MballFrame) frame).showBuffer();
			return true;
		case '8':
			wedge8ModePuzzle();
			break;
		case '4':
			wedge4ModePuzzle();
			break;
		case '6':
			wedge6ModePuzzle();
			break;
		case '2':
			wedge2ModePuzzle();
			break;
		case '0':
			wedge10ModePuzzle();
			break;
		case '=':
			wedge12ModePuzzle();
			break;
		case 'g':
			getPuzzle();
			break;
		case 'w':
			writePuzzle();
			break;
		case 'G':
			getPuzzleFile();
			break;
		case 'W':
			writePuzzleFile();
			break;
		case 'u':
		case 'U':
			undoPuzzle();
			break;
		case 'r':
		case 'R':
			redoPuzzle();
			break;
		case 'c':
		case 'C':
			clearPuzzle();
			break;
		case 'z':
		case 'Z':
			randomizePuzzle();
			break;
		case 'p':
		case 'P':
			practicePuzzle();
			break;
		case 's':
		case 'S':
			solvePuzzle();
			break;
		case 'o':
		case 'O':
			orientizePuzzle();
			break;
		case 'i':
		case 'I':
			incrementPuzzle();
			break;
		case 'd':
		case 'D':
			decrementPuzzle();
			break;
		case 't':
		case 'T':
			perspectivePuzzle();
			break;
		case 'q':
		case 'Q':
		case 'x':
		case 'X':
		case '\03': /* Control-C */
			((MballFrame) frame).quit();
			break;
		default:
			return false;
		}
		return true;
	}

	void processPuzzleInput(int x, int y, int direction, boolean control) {
		movePuzzleInput(x, y, direction, control);
	}

	@Override
	public void mousePressed(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (scanPaint || stackPaint || mouseDown)
			return;
		if (((modifiers & InputEvent.META_DOWN_MASK) == 0) &&
				((modifiers & InputEvent.ALT_DOWN_MASK) == 0)) {
			if (randomPaints == 0) {
				selectPuzzle((modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
					event.getX(), event.getY());
				mouseDown = true;
			}
		} else {
			if ((event.getClickCount() & 1) == 1)
				randomizePuzzleWithQuery();
			else
				randomizePuzzle();
		}
	}

	@Override
	public void mouseClicked(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (mouseDown) { // Do not care which mouse button
			releasePuzzle((modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				event.getX(), event.getY());
			mouseDown = false;
		}
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		currentForeground = foreground;
		framePaint = true;
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseExited(MouseEvent event) {
		if (mouseDown && currentSector >= 0) {
			releasePaint = true;
		}
		currentForeground = borderColor;
		framePaint = true;
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseDragged(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		lastX = event.getX();
		lastY = event.getY();
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent event) {
		int notches = event.getWheelRotation();
		boolean control = ((event.getModifiersEx() &
			InputEvent.CTRL_DOWN_MASK) != 0);

		processPuzzleInput(lastX, lastY,
			((notches < 0) ? TOP : BOTTOM), control);
	}

	@Override
	public void keyPressed(KeyEvent event) {
		boolean control = ((event.getModifiersEx() &
			InputEvent.CTRL_DOWN_MASK) != 0);

		switch (event.getKeyCode()) {
		case KeyEvent.VK_ESCAPE:
			((MballFrame) frame).shuffleDown();
			return;
		case KeyEvent.VK_NUM_LOCK:
			processPuzzleInput(lastX, lastY, TTL, control);
			return;
		case KeyEvent.VK_F11:
		case KeyEvent.VK_DIVIDE:
			processPuzzleInput(lastX, lastY, CCW, control);
			return;
		case KeyEvent.VK_MULTIPLY:
			processPuzzleInput(lastX, lastY, TTR, control);
			return;
		case KeyEvent.VK_NUMPAD7:
		case KeyEvent.VK_HOME:
			processPuzzleInput(lastX, lastY, TL, control);
			return;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
		case KeyEvent.VK_NUMPAD8:
			processPuzzleInput(lastX, lastY, TOP, control);
			return;
		case KeyEvent.VK_PAGE_UP:
		case KeyEvent.VK_NUMPAD9:
			processPuzzleInput(lastX, lastY, TR, control);
			return;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
		case KeyEvent.VK_NUMPAD4:
			processPuzzleInput(lastX, lastY, LEFT, control);
			return;
		case KeyEvent.VK_F12:
		case KeyEvent.VK_CLEAR:
		case KeyEvent.VK_NUMPAD5:
			processPuzzleInput(lastX, lastY, CW, control);
			return;
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
		case KeyEvent.VK_NUMPAD6:
			processPuzzleInput(lastX, lastY, RIGHT, control);
			return;
		case KeyEvent.VK_END:
		case KeyEvent.VK_NUMPAD1:
			processPuzzleInput(lastX, lastY, BL, control);
			return;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
		case KeyEvent.VK_NUMPAD2:
			processPuzzleInput(lastX, lastY, BOTTOM, control);
			return;
		case KeyEvent.VK_PAGE_DOWN:
		case KeyEvent.VK_NUMPAD3:
			processPuzzleInput(lastX, lastY, BR, control);
			return;
		case KeyEvent.VK_INSERT:
		case KeyEvent.VK_NUMPAD0:
			processPuzzleInput(lastX, lastY, BBL, control);
			return;
		case KeyEvent.VK_DECIMAL:
		case KeyEvent.VK_DELETE:
			processPuzzleInput(lastX, lastY, BBR, control);
			return;
		default:
			processKey(event.getKeyChar());
		}
	}

	@Override
	public void keyReleased(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	@Override
	public void keyTyped(KeyEvent event) {
		/* Method required but keyPressed does work */
	}
}
