package com.sillycycle.bagleyd.mball;

/*
 * @(#)MballStack.java
 *
 * Copyright 1994 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Deque;
import java.util.LinkedList;

/**
 * The <code>MballStack</code> class allows one to undo and redo
 * in the <code>MballFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class MballStack {
	//int direction, control;
	private int packed; /* This makes assumptions on the data. */
	private int sector; /* Do not make assumptions on this one. */
	private int count;
	private final Deque<String> stack = new LinkedList<>();

	private void writeMoveDirection(int direction) {
		//this.direction = direction;
		packed += ((direction & 0xFF) << 4);
	}

	private void writeMoveControl(int control) {
		//this.control = control;
		packed += control & 0xF;
	}

	private void writeMoveSector(int sect) {
		this.sector = sect;
	}

	int readMoveDirection() {
		//return this.direction;
		return ((packed >> 4) & 0xFF);
	}

	int readMoveControl() {
		//return this.control;
		return (packed & 0xF);
	}

	int readMoveSector() {
		return this.sector;
	}

	void setMove(int direction, int control, int sector) {
		packed = 0;
		writeMoveDirection(direction);
		writeMoveControl(control);
		writeMoveSector(sector);
		stack.push(Integer.toString(packed));
		stack.push(Integer.toString(sector));
		count++;
	}

	void getMove() {
		sector = Integer.parseInt(stack.pop());
		packed = Integer.parseInt(stack.pop());
		count--;
	}

	boolean madeMoves() {
		return !stack.isEmpty();
	}

	void flushMoves() {
		while (madeMoves())
			getMove();
		count = 0; // probably do not need
	}

	int numMoves() {
		return count;
	}
}
